#include "main.h"

static Scene *s; 
static Matrix4 nm, mv, mvp, mdpy;
static RContext *rc;
static void *vdata;


int main(int argc, char **argv)
{
  Object *o;
  iMesh *im; iTri *t;
  Hpoly *pl;

  init_lang();
  s = scene_read();
  s->objs = graph_flatten(graph_transform(s->objs));
  init_render(s);
  for (o = s->objs; o != NULL; o = o->next) {
    im = tesselate(o);
    for (t = im->t; t != NULL; t = t->next, hplist_free(pl)) {
      pl = assembly(o, t);
      if (!vertex_proc(pl, vdata, o)) 
	continue;
      if (poly_clip(pl, efree, 1)) {
	hpoly_transform(hpoly_homo(pl), mdpy);
	scan_hpoly(pl, pixel_proc, vdata);
      }
    }
  }
  raster_proc(s);
  exit(0);
}


void init_lang(void)
{
  lang_defun("scene", scene_parse);
  lang_defun("view", view_parse);
  lang_defun("dist_light", distlight_parse);
  lang_defun("plastic", plastic_parse);

  lang_defun("primobj", obj_parse);
  lang_defun("sphere", sphere_parse);

  lang_defun("meshobj", obj_parse);
  lang_defun("mesh", imesh_parse);

  lang_defun("group", group_parse);
  lang_defun("affine", xaffine_parse);

#ifdef MAP_SHADE
  lang_defun("slideproj", slideproj_parse);
  lang_defun("shiny", shiny_parse);
  lang_defun("rough", rough_parse);
  lang_defun("textured", textured_parse);
  lang_defun("imagemap", imagemap_parse);
  lang_defun("chequer", chequer_parse);
#endif

  lang_defun("arg", arg_parse);
  lang_defun("motor", motor_parse);

  lang_defun("include", include_parse);
}



/* ------------------ stage ops  -------------------- */

void init_render(Scene *s)
{
  mv = s->view->V; nm = s->view->Vinv;
  mvp =  m4_m4prod(m4_m4prod(s->view->P, s->view->C), s->view->V);
  mdpy = s->view->S;
  zbuf_init(s->img->w, s->img->h);
  rc_sset(rc = NEWSTRUCT(RContext), s->view, s->lights, s->img);
  vdata = (void *) NEWSTRUCT(SHADE_DATA);
}

iMesh *tesselate(Object *obj)
{
  switch (obj->type) {
  case V_PRIM: return prim_uv_decomp(prim_(obj), 1.);
  case V_TRIMESH: return mesh_(obj);
  }
  return NULL;
}

Hpoly *assembly(Object *obj, iTri *t)
{
  switch (obj->type) {
  case V_PRIM: return prim_polys(prim_(obj), t);
  case V_TRIMESH: return mesh_polys(mesh_(obj), t);
  }
  return NULL;
}

int vertex_proc(Hpoly *p, void *d, Object *obj)
{
  if (is_backfacing(p, v3_sub(hpoly_centr(p), s->view->center)))
    return FALSE;
  SHADE_PREP(SHADE_PREP_PARAMS);
  hpoly_transform(p, mvp);
#ifdef MAP_SHADE
  poly_wz_hpoly(WL(p), p);
  texture_wscale(WL(p), TL(p));
#endif
  d = SHADE_SET(SHADE_SET_PARAMS);
  return TRUE;
}

void pixel_proc(Vector4 v,int n,int lv,Real lt,int rv,Real rt,Real st,void *c)
{
  if (zbuf_store(v))
    SHADE_PAINT(v, n, lv, lt, rv, rt, st, c);
}

void raster_proc(Scene *s)
{
  img_write(s->img, "stdout", 0);
}


/* -------------------- obj dependent ----------------------- */

void shade_dummy(int i) {}

Hpoly *prim_polys(Prim *s, iTri *t)
{
  int i;
  Hpoly *pl = hplist_alloc(SHADE_NPOLYS);

  for (i = 0; i < 3; i++) {
    Vector3 p = imesh_mpoint(t->m, t->v[i]);
    PL(pl)->v[i] = v4_v3conv(prim_point(s, p.x, p.y));
#if defined(GOURAUD_SHADE) || defined(PHONG_SHADE) || defined(MAP_SHADE)
    NL(pl)->v[i] = v4_v3conv(prim_normal(s, p.x, p.y));
#endif
#ifdef GOURAUD_SHADE
    CL(pl)->v[i] = v4_make(0,0,0,0);
#endif
#if defined(PHONG_SHADE) || defined(MAP_SHADE)
    SL(pl)->v[i] = PL(pl)->v[i];
#endif
#ifdef MAP_SHADE
    WL(pl)->v[i] = PL(pl)->v[i];
    TL(pl)->v[i] = v4_v3conv(prim_texc(s, p.x, p.y));
    DUL(pl)->v[i] = v4_v3conv(v3_unit(prim_du(s, p.x, p.y)));
    DVL(pl)->v[i] = v4_v3conv(v3_unit(prim_dv(s, p.x, p.y)));
#endif
  }
  TAIL(pl) = NULL;
  return pl;
}

Hpoly *mesh_polys(iMesh *m, iTri *t)
{
  int i;
  Hpoly *pl = hplist_alloc(SHADE_NPOLYS);

  for (i = 0; i < 3; i++) {
    PL(pl)->v[i] = v4_v3conv(imesh_point(m, t->v[i]));
#if defined(GOURAUD_SHADE) || defined(PHONG_SHADE) || defined(MAP_SHADE)
    NL(pl)->v[i] = v4_v3conv(imesh_normal(m, t->v[i]));
#endif
#ifdef GOURAUD_SHADE
    CL(pl)->v[i] = v4_make(0,0,0,0);
#endif
#if defined(PHONG_SHADE) || defined(MAP_SHADE)
    SL(pl)->v[i] = PL(pl)->v[i];
#endif
#ifdef MAP_SHADE
    WL(pl)->v[i] = PL(pl)->v[i];
    TL(pl)->v[i] = v4_v3conv(imesh_texc(m, t->v[i]));
    DUL(pl)->v[i] = v4_v3conv(imesh_du(m, t->v[i]));
    DVL(pl)->v[i] = v4_v3conv(imesh_dv(m, t->v[i]));
#endif
  }
  TAIL(pl) = NULL;
  return pl;
}

